#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2008 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ServiceModel;
using HeuristicLab.Hive.Contracts.Interfaces;
using HeuristicLab.Hive.Slave.Common;
using HeuristicLab.PluginInfrastructure;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using HeuristicLab.Tracing;
using HeuristicLab.Hive.Contracts;
using HeuristicLab.Hive.Contracts.BusinessObjects;
using HeuristicLab.Hive.Slave.Communication.SlaveService;
using HeuristicLab.Hive.Contracts.ResponseObjects;
using HeuristicLab.Hive.Slave.Communication.Properties;

namespace HeuristicLab.Hive.Slave.Communication {

  /// <summary>
  /// WcfService class is implemented as a Singleton and works as a communication Layer with the Server
  /// </summary>
  public class WcfService {
    private static WcfService instance;
    /// <summary>
    /// Getter for the Instance of the WcfService
    /// </summary>
    /// <returns>the Instance of the WcfService class</returns>
    public static WcfService Instance {
      get {
        if (instance == null) {
          Logger.Debug("New WcfService Instance created");
          instance = new WcfService();
        }
        return instance;
      }
    }

    public DateTime ConnectedSince { get; private set; }
    public NetworkEnum.WcfConnState ConnState { get; private set; }
    public bool LoggedIn { get; set; }

    private string serverIp;
    public string ServerIp {
      get { return serverIp; }
      set {
        if (serverIp != value) {
          serverIp = value;
          if (ServerChanged != null)
            ServerChanged(this, new EventArgs());
        }
      }
    }

    public event EventHandler ConnectionRestored;
    public event EventHandler ServerChanged;
    public event EventHandler Connected;

    /// <summary>
    /// Constructor
    /// </summary>
    private WcfService() {
      ConnState = NetworkEnum.WcfConnState.Disconnected;
      LoggedIn = false;
    }

    /// <summary>
    /// Connects with the Server, registers the events and fires the Connected (and quiet possibly the ConnectionRestored) Event.
    /// </summary>
    public void Connect() {
      SlaveService.ISlaveFacade client = null;
      try {
        Logger.Debug("Starting the Connection Process");
        if (String.Empty.Equals(ServerIp)) {
          Logger.Info("No Server IP set!");
          return;
        }

        Logger.Debug("Creating the new connection proxy");
        client = ServiceLocator.CreateStreamedSlaveFacade(ServerIp);
        Logger.Debug("Created the new connection proxy");

        ConnState = NetworkEnum.WcfConnState.Connected;
        ConnectedSince = DateTime.Now;

        if (Connected != null) {
          Connected(this, new EventArgs());
        }
        if (ConnState == NetworkEnum.WcfConnState.Failed) {
          ConnectionRestored(this, new EventArgs());
        }
      }
      catch (Exception ex) {
        HandleNetworkError(ex);
      }
      finally {
        ServiceLocator.DisposeSlaveClient(client);
      }
    }

    ///// <summary>
    ///// Disconnects the Slave from the Server
    ///// </summary>
    public void Disconnect() {
      ConnState = NetworkEnum.WcfConnState.Disconnected;
      LoggedIn = false;
    }

    /// <summary>
    /// Network communication Error Handler - Every network error gets logged and the connection switches to faulted state
    /// </summary>
    /// <param name="e">The Exception</param>
    private void HandleNetworkError(Exception e) {
      ConnState = NetworkEnum.WcfConnState.Failed;
      LoggedIn = false;
      Logger.Error("Network exception occurred: " + e);
    }

    /// <summary>
    /// Methods for the Server Login
    /// </summary>
    public void Login(SlaveDto slaveInfo) {
      SlaveService.ISlaveFacade client = null;
      try {
        if (ConnState == NetworkEnum.WcfConnState.Connected) {
          Logger.Debug("STARTED: Login Sync");
          client = ServiceLocator.CreateSlaveFacade(ServerIp);
          Response res = client.Login(slaveInfo);
          if (res.StatusMessage != ResponseStatus.Ok) {
            Logger.Error("FAILED: Login Failed! " + res.StatusMessage);
            throw new Exception(res.StatusMessage.ToString());
          } else {
            Logger.Info("ENDED: Login succeeded" + res.StatusMessage);
            LoggedIn = true;
          }
        }
      }
      catch (Exception e) {
        HandleNetworkError(e);
      }
      finally {
        ServiceLocator.DisposeSlaveClient(client);
      }
    }

    /// <summary>
    /// Pull a Job from the Server
    /// </summary>
    #region PullJob
    public event System.EventHandler<GetJobCompletedEventArgs> GetJobCompleted;
    public void GetJobAsync(Guid guid) {
      if (LoggedIn) {
        Logger.Debug("STARTED: Fetching of Jobs from Server for Slave");
        SlaveService.ISlaveFacade client = ServiceLocator.CreateStreamedSlaveFacade(ServerIp);
        //client.GetStreamedJobAsync(guid);
        client.BeginGetStreamedJob(guid, (ar => {
          if (ar.IsCompleted) {
            Stream stream = null;
            MemoryStream memStream = null;
            try {
              Logger.Debug("ENDED: Fetching of Jobs from Server for Slave");
              stream = client.EndGetStreamedJob(ar);

              //first deserialize the response
              BinaryFormatter formatter = new BinaryFormatter();
              ResponseObject<JobDto> response = (ResponseObject<JobDto>)formatter.Deserialize(stream);

              //second deserialize the BLOB
              memStream = new MemoryStream();

              byte[] buffer = new byte[3024];
              int read = 0;
              while ((read = stream.Read(buffer, 0, buffer.Length)) > 0) {
                memStream.Write(buffer, 0, read);
              }

              memStream.Close();

              GetJobCompletedEventArgs completedEventArgs = new GetJobCompletedEventArgs(new object[] { response, memStream.GetBuffer() }, null, !ar.IsCompleted, ar.AsyncState);
              GetJobCompleted(this, completedEventArgs);
            }
            catch (Exception ex) {
              Logger.Error(ex);
            }
            finally {
              if (stream != null)
                stream.Dispose();

              if (memStream != null)
                memStream.Dispose();
            }
          } else
            HandleNetworkError(new FaultException("GetJobAsync did not complete"));

          ServiceLocator.DisposeSlaveClient(client);
        }), null);
      }
    }

    #endregion

    /// <summary>
    /// Send back finished Job Results
    /// </summary>
    #region SendJobResults
    public event System.EventHandler<StoreFinishedJobResultCompletedEventArgs> GetFinishedJobResultCompleted;
    public void GetFinishedJobResultAsync(Guid clientId, Guid jobId, byte[] result, double percentage, Exception exception, bool finished) {
      if (LoggedIn) {
        Logger.Debug("STARTED: Sending back the finished job results");
        Logger.Debug("Building stream");
        Stream stream = GetStreamedJobResult(clientId, jobId, result, percentage, exception);
        Logger.Debug("Builded stream");
        Logger.Debug("Making the call");
        //proxy.StoreFinishedJobResultStreamedAsync(stream, stream);
        SlaveService.ISlaveFacade client = ServiceLocator.CreateStreamedSlaveFacade(ServerIp);
        client.BeginStoreFinishedJobResultStreamed(stream, (ar => {
          Logger.Debug("Finished storing the job");
          if (stream != null)
            stream.Dispose();

          if (ar.IsCompleted) {
            var res = client.EndStoreFinishedJobResultStreamed(ar);
            StoreFinishedJobResultCompletedEventArgs args = new StoreFinishedJobResultCompletedEventArgs(new object[] { res }, null, false, null);
            Logger.Debug("calling the Finished Job Event");
            GetFinishedJobResultCompleted(this, args);
            Logger.Debug("ENDED: Sending back the finished job results");
          } else {
            HandleNetworkError(new FaultException("GetFinishedJobResultAsync did not complete"));
          }
          ServiceLocator.DisposeSlaveClient(client);
        }), null);
      }
    }

    #endregion

    #region Processsnapshots
    public event System.EventHandler<ProcessSnapshotCompletedEventArgs> ProcessSnapshotCompleted;
    public void ProcessSnapshotAsync(Guid clientId, Guid jobId, byte[] result, double percentage, Exception exception, bool finished) {
      if (LoggedIn) {
        Stream stream = GetStreamedJobResult(clientId, jobId, result, percentage, exception);
        //proxy.ProcessSnapshotStreamedAsync(stream, stream);
        SlaveService.ISlaveFacade client = ServiceLocator.CreateStreamedSlaveFacade(ServerIp);
        client.BeginProcessSnapshotStreamed(stream, (ar => {
          if (stream != null)
            stream.Dispose();

          if (ar.IsCompleted) {
            var res = client.EndStoreFinishedJobResultStreamed(ar);
            ProcessSnapshotCompletedEventArgs args = new ProcessSnapshotCompletedEventArgs(new object[] { res }, null, false, null);
            ProcessSnapshotCompleted(this, args);
          } else {
            HandleNetworkError(new FaultException("ProcessSnapshotAsync did not complete"));
          }
          ServiceLocator.DisposeSlaveClient(client);
        }), null);
      }
    }

    #endregion

    /// <summary>
    /// Methods for sending the periodically Heartbeat
    /// </summary>
    #region Heartbeat

    public event System.EventHandler<ProcessHeartBeatCompletedEventArgs> ProcessHeartBeatCompleted;
    public void ProcessHeartBeatAsync(HeartBeatData hbd) {
      if (LoggedIn) {
        Logger.Debug("STARTING: sending heartbeat");
        //proxy.ProcessHeartBeatAsync(hbd);
        SlaveService.ISlaveFacade client = ServiceLocator.CreateStreamedSlaveFacade(ServerIp);
        client.BeginProcessHeartBeat(hbd, (ar => {
          if (ar.IsCompleted) {
            var res = client.EndProcessHeartBeat(ar);
            if (res.StatusMessage == ResponseStatus.Ok) {
              ProcessHeartBeatCompleted(this, new ProcessHeartBeatCompletedEventArgs(new object[] { res }, null, false, null));
              Logger.Debug("ENDED: sending heartbeats");
            } else {
              Logger.Error("FAILED: sending heartbeats: " + res.StatusMessage.ToString());
            }
          } else {
            HandleNetworkError(new FaultException("ProcessHeartBeatAsync did not complete"));
          }
          ServiceLocator.DisposeSlaveClient(client);
        }), null);
      }
    }

    #endregion

    /// <summary>
    /// Send back finished and Stored Job Results
    /// </summary>
    private Stream GetStreamedJobResult(Guid clientId, Guid jobId, byte[] result, double percentage, Exception exception) {
      JobResult jobResult = new JobResult();
      jobResult.SlaveId = clientId;
      jobResult.JobId = jobId;
      jobResult.Percentage = percentage;
      jobResult.Exception = exception != null ? exception.Message : "";

      MultiStream stream = new MultiStream();

      //first send result
      stream.AddStream(new StreamedObject<JobResult>(jobResult));

      //second stream the job binary data
      MemoryStream memStream = new MemoryStream(result, false);
      stream.AddStream(memStream);

      return stream;
    }

    public ResponseResultReceived StoreFinishedJobResultsSync(Guid clientId, Guid jobId, byte[] result, double percentage, Exception exception, bool finished) {
      SlaveService.ISlaveFacade client = ServiceLocator.CreateStreamedSlaveFacade(ServerIp);
      ResponseResultReceived res = client.StoreFinishedJobResultStreamed(GetStreamedJobResult(clientId, jobId, result, percentage, exception));
      ServiceLocator.DisposeSlaveClient(client);
      return res;
    }

    public Response IsJobStillNeeded(Guid jobId) {
      try {
        Logger.Debug("STARTING: Sync call: IsJobStillNeeded");
        SlaveService.ISlaveFacade client = ServiceLocator.CreateStreamedSlaveFacade(ServerIp);
        Response res = client.IsJobStillNeeded(jobId);
        ServiceLocator.DisposeSlaveClient(client);
        Logger.Debug("ENDED: Sync call: IsJobStillNeeded");
        return res;
      }
      catch (Exception e) {
        HandleNetworkError(e);
        return null;
      }
    }

    public ResponseResultReceived ProcessSnapshotSync(Guid clientId, Guid jobId, byte[] result, double percentage, Exception exception) {
      try {
        SlaveService.ISlaveFacade client = ServiceLocator.CreateStreamedSlaveFacade(ServerIp);
        var res = client.ProcessSnapshotStreamed(GetStreamedJobResult(clientId, jobId, result, percentage, exception));
        ServiceLocator.DisposeSlaveClient(client);
        return res;
      }
      catch (Exception e) {
        HandleNetworkError(e);
        return null;
      }
    }

    public IEnumerable<CachedHivePluginInfoDto> RequestPlugins(List<HivePluginInfoDto> requestedPlugins) {
      try {
        Logger.Debug("STARTED: Requesting Plugins for Job");
        Logger.Debug("STARTED: Getting the stream");
        SlaveService.ISlaveFacade client = ServiceLocator.CreateStreamedSlaveFacade(ServerIp);
        Stream stream = client.GetStreamedPlugins(requestedPlugins.ToArray());
        ServiceLocator.DisposeSlaveClient(client);
        Logger.Debug("ENDED: Getting the stream");
        BinaryFormatter formatter = new BinaryFormatter();
        Logger.Debug("STARTED: Deserializing the stream");
        ResponseList<CachedHivePluginInfoDto> response = (ResponseList<CachedHivePluginInfoDto>)formatter.Deserialize(stream);
        Logger.Debug("ENDED: Deserializing the stream");
        if (stream != null)
          stream.Dispose();
        return response.List;
      }
      catch (Exception e) {
        HandleNetworkError(e);
        return null;
      }
    }

    public void Logout(Guid guid) {
      try {
        Logger.Debug("STARTED: Logout");
        SlaveService.ISlaveFacade client = ServiceLocator.CreateStreamedSlaveFacade(ServerIp);
        client.Logout(guid);
        ServiceLocator.DisposeSlaveClient(client);
        Logger.Debug("ENDED: Logout");
      }
      catch (Exception e) {
        HandleNetworkError(e);
      }
    }

    public ResponseCalendar GetCalendarSync(Guid clientId) {
      try {
        Logger.Debug("STARTED: Syncing Calendars");
        SlaveService.ISlaveFacade client = ServiceLocator.CreateStreamedSlaveFacade(ServerIp);
        ResponseCalendar cal = client.GetCalendar(clientId);
        ServiceLocator.DisposeSlaveClient(client);
        Logger.Debug("ENDED: Syncing Calendars");
        return cal;
      }
      catch (Exception e) {
        HandleNetworkError(e);
        return null;
      }
    }

    public Response SetCalendarStatus(Guid clientId, CalendarState state) {
      try {
        Logger.Debug("STARTED: Setting Calendar status to: " + state);
        SlaveService.ISlaveFacade client = ServiceLocator.CreateStreamedSlaveFacade(ServerIp);
        Response resp = client.SetCalendarStatus(clientId, state);
        ServiceLocator.DisposeSlaveClient(client);
        Logger.Debug("ENDED: Setting Calendar status to: " + state);
        return resp;
      }
      catch (Exception e) {
        HandleNetworkError(e);
        return null;
      }
    }
  }
}